__package__ = 'weipyweb'
__description__ = 'Beautify the original weipyweb'
__url__ = 'https://github.com/acloudtwei'
__version__ = "0.0.1"
__version_info__ = (1, 5, 2, 0)
__author__ = 'Acloudtwei'
__author_email__ = 'acloudtwei@gmail.com'
__license__ = 'MIT'
